// Copyright 2019 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// Package prototext marshals and unmarshals protocol buffer messages as the
// textproto format.
package prototext
